import requests
from dockyter.backend import APIBackend


class FakeResponse:
    def __init__(self, ok=True, status_code=200, json_data=None):
        self.ok = ok
        self.status_code = status_code
        self._json_data = json_data or {}

    def raise_for_status(self):
        if not self.ok:
            raise Exception(f"HTTP {self.status_code}")

    def json(self):
        return self._json_data


def test_api_backend_success(monkeypatch):
    FAKE_API_GOOD_RESULT = "Hello from fake API"
    calls_get = []
    calls_post = []

    def fake_get(url, timeout):
        calls_get.append(url)
        assert url.endswith("/health")
        return FakeResponse(ok=True, status_code=200)

    def fake_post(url, json, timeout):
        calls_post.append((url, json))
        assert url.endswith("/execute")
        assert json["cmd"] == "echo hello"
        assert "ubuntu:latest" in json["args"]
        return FakeResponse(
            ok=True,
            status_code=200,
            json_data={
                "stdout": FAKE_API_GOOD_RESULT,
                "stderr": ""
            },
        )

    monkeypatch.setattr(requests, "get", fake_get)
    monkeypatch.setattr(requests, "post", fake_post)

    backend = APIBackend("https://api.example.com")
    result = backend.dockyter_command("echo hello", "ubuntu:latest -v /tmp:/tmp")

    assert result.stdout == FAKE_API_GOOD_RESULT
    assert result.stderr == ""

    assert len(calls_get) == 1
    assert len(calls_post) == 1


def test_api_backend_unreachable(monkeypatch):
    calls_get = []
    calls_post = []

    def fake_get(url, timeout):
        calls_get.append(url)
        raise requests.RequestException("Connection error")

    def fake_post(url, json, timeout):
        calls_post.append((url, json))
        return FakeResponse(
            ok=True, 
            status_code=200, 
            json_data={
                "stdout": "SHOULD NOT BE USED",
                "stderr": "SHOULD NOT BE USED"
            }
        )

    monkeypatch.setattr(requests, "get", fake_get)
    monkeypatch.setattr(requests, "post", fake_post)

    backend = APIBackend("https://api.example.com")
    result = backend.dockyter_command("echo hello", "ubuntu:latest -v /tmp:/tmp")

    assert result.stdout == ""
    assert "API backend unreachable at " in result.stderr
    assert len(calls_get) == 1
    assert len(calls_post) == 0


def test_api_backend_forbidden_flag(monkeypatch):
    calls_get = []
    calls_post = []

    def fake_get(url, timeout):
        calls_get.append(url)
        return FakeResponse(ok=True, status_code=200)

    def fake_post(url, json, timeout):
        calls_post.append((url, json))
        return FakeResponse(
            ok=True,
            status_code=200,
            json_data={
                "stdout": "SHOULD NOT BE USED", 
                "stderr": "SHOULD NOT BE USED"
            },
        )

    monkeypatch.setattr(requests, "get", fake_get)
    monkeypatch.setattr(requests, "post", fake_post)

    backend = APIBackend("https://api.example.com")
    result = backend.dockyter_command("echo hello", "ubuntu:latest --privileged")

    assert result.stdout == ""
    assert "Forbidden Docker flag detected: " in result.stderr

    assert len(calls_get) == 1
    assert len(calls_post) == 0



from operator import call
import subprocess
from dockyter.backend import DockerBackend, validate_docker_args

def test_validate_docker_args():
    VALID_ARGS = ["ubuntu:latest -v /tmp:/tmp", "ubuntu:latest -e VAR=1"]
    INVALID_ARGS = ["ubuntu:latest --privileged", "ubuntu:latest --network=host", "ubuntu:latest --net=host"]
    
    for args in VALID_ARGS:
        ok, bad_flag_message = validate_docker_args(args)
        assert ok
        assert bad_flag_message == ""
    
    for args in INVALID_ARGS:
        ok, bad_flag_message = validate_docker_args(args)
        assert not ok
        assert "Forbidden Docker flag detected: " in bad_flag_message

def test_docker_backend_sucess(monkeypatch):
    FAKE_DOCKER_GOOD_RESULT = "Hello from fake docker"
    calls_run = []
    calls_get_status = []

    def fake_run(cmd, *args, **kwargs):
        calls_run.append(cmd)
        return subprocess.CompletedProcess(
            args=cmd,
            returncode=0,
            stdout=FAKE_DOCKER_GOOD_RESULT,
            stderr="",
        )

    def fake_get_status(self):
        calls_get_status.append(True)
        return True, "Docker is installed and the daemon is running."

    monkeypatch.setattr(subprocess, "run", fake_run)
    monkeypatch.setattr(DockerBackend, "get_status", fake_get_status)

    backend = DockerBackend()
    
    result = backend.dockyter_command("echo hello", "ubuntu:latest -v /tmp:/tmp")

    assert result.stdout == FAKE_DOCKER_GOOD_RESULT
    assert result.stderr == ""

    assert len(calls_get_status) == 1
    assert len(calls_run) == 1
    full_cmd = calls_run[0]

    assert full_cmd[:3] == ["docker", "run", "--rm"]
    assert "ubuntu:latest" in full_cmd
    assert "bash" in full_cmd
    assert "-lc" in full_cmd
    assert "echo hello" in full_cmd

def test_docker_backend_docker_not_installed(monkeypatch):
    DOCKER_NOT_INSTALLED_MESSAGE = "Docker is not installed or not available in the system PATH."
    calls_run = []
    calls_get_status = []

    def fake_run(cmd, *args, **kwargs):
        calls_run.append(cmd)
        return subprocess.CompletedProcess(
            args=cmd,
            returncode=1,
            stdout="SHOULD NOT BE USED",
            stderr="SHOULD NOT BE USED",
        )

    def fake_get_status(self):
        calls_get_status.append(True)
        return False, DOCKER_NOT_INSTALLED_MESSAGE

    monkeypatch.setattr(subprocess, "run", fake_run)
    monkeypatch.setattr(DockerBackend, "get_status", fake_get_status)

    backend = DockerBackend()
    result = backend.dockyter_command("echo hello", "ubuntu:latest -v /tmp:/tmp")

    assert result.stdout == ""
    assert result.stderr == DOCKER_NOT_INSTALLED_MESSAGE

    assert len(calls_get_status) == 1
    assert len(calls_run) == 0  

def test_docker_backend_forbidden_flag(monkeypatch):
    calls_run = []
    calls_get_status = []

    def fake_run(cmd, *args, **kwargs):
        calls_run.append(cmd)
        return subprocess.CompletedProcess(
            args=cmd,
            returncode=0,
            stdout="SHOULD NOT BE USED",
            stderr="SHOULD NOT BE USED",
        )

    def fake_get_status(self):
        calls_get_status.append(True)
        return True, "Docker is installed and the daemon is running."

    monkeypatch.setattr(subprocess, "run", fake_run)
    monkeypatch.setattr(DockerBackend, "get_status", fake_get_status)

    backend = DockerBackend()
    result = backend.dockyter_command("echo hello", "ubuntu:latest --privileged")

    assert result.stdout == ""
    assert "Forbidden Docker flag detected: " in result.stderr

    assert len(calls_get_status) == 1
    assert len(calls_run) == 0



from IPython.core.magic import (Magics, magics_class, line_magic, cell_magic)
from IPython.core.getipython import get_ipython
from .backend import DockerBackend, APIBackend, CommandResult

BACKEND_MODE_DOCKER = "docker"
BACKEND_MODE_API = "api"

@magics_class
class Dockyter(Magics):
    def __init__(self, shell=None, **kwargs):
        super().__init__(shell=shell, **kwargs)
        self.docker_args: str = ""
        self.original_system = None
        self.docker_reroute_enabled = False
        self.backend = DockerBackend()

    def print_error(self, message: str):
        print(f"\033[91m{message}\033[0m")

    def print_outputs(self, result: CommandResult):
        has_stdout = bool(result.stdout)
        has_stderr = bool(result.stderr)

        if has_stdout:
            print(result.stdout)

        if has_stderr:
            if not has_stdout:
                self.print_error(result.stderr)
            else:
                print(result.stderr)


    def docker_console(self, cmd):
        result = self.backend.dockyter_command(cmd, self.docker_args)
        self.print_outputs(result)

    @line_magic("docker_status")
    def docker_status_magic(self, line: str = ""):
        ok, docker_status = self.backend.get_status()
        docker_backend_type = "Docker" if isinstance(self.backend, DockerBackend) else "API"

        print("=== Dockyter Backend Status ===")
        print(f"Dockyter backend type: {docker_backend_type}")
        print(f"Dockyter backend status: {'available' if ok else 'unavailable'}")
        print(f"Dockyter backend status: {docker_status}")
        print(f"Dockyer redirection for '!': {'on' if self.docker_reroute_enabled else 'off'}")
        
        if self.docker_args:
            print(f"Current docker args: {self.docker_args}")

    @line_magic("docker_backend")
    def docker_backend_magic(self, line: str):
        parts = line.strip().split()

        if not parts:
            self.print_error("Usage: %docker_backend <docker|api> [api_url]")
            return

        mode = parts[0].lower()

        if mode == BACKEND_MODE_DOCKER:
            self.backend = DockerBackend()
            print("Switched backend to Docker (local daemon).")
            self.docker_status_magic()
            return

        if mode == BACKEND_MODE_API:
            if len(parts) < 2:
                self.print_error("Usage: %docker_backend api <api_url>")
                return
            url = parts[1]
            self.backend = APIBackend(url)
            print(f"Switched backend to APIBackend ({url}).")
            self.docker_status_magic()
            return

        self.print_error(f"Unknown backend. Use {BACKEND_MODE_DOCKER} or {BACKEND_MODE_API}.")


    @line_magic("docker")
    def docker_line_magic(self, line: str):
        self.docker_args = line

        ip = get_ipython()
        if ip is not None:
            self.original_system = ip.system
            self.docker_on_magic()
            result = self.backend.dockyter_command("echo 'Connected'", args=line)
            self.print_outputs(result)
        else:
            self.print_error("Could not access IPython instance to reroute '!' commands.")

    @line_magic("docker_on")
    def docker_on_magic(self, line: str = ""):
        ip = get_ipython()
        if ip is not None:
            self.original_system = ip.system
            ip.system = self.docker_console
            self.docker_reroute_enabled = True
        else:
            self.print_error("Could not access IPython instance to reroute '!' commands.")

    @line_magic("docker_off")
    def docker_off_magic(self, line: str = ""):
        ip = get_ipython()
        if ip is not None and self.original_system is not None:
            ip.system = self.original_system
            self.docker_reroute_enabled = False
        else:
            self.print_error("Could not access IPython instance to restore original '!' behavior.")

    @cell_magic("docker")
    def docker_cell_magic(self, line: str, cell):
        result = self.backend.dockyter_command(cell, args=line)
        self.print_outputs(result)



import nbformat
import subprocess
import pytest
from pathlib import Path
from nbconvert.preprocessors import ExecutePreprocessor

RED_ANSI = "\x1b[91m"
RESET_ANSI = "\x1b[0m"
NOTEBOOK_DIR = Path("docs/exemples")
API_URL = "http://127.0.0.1:8000"

@pytest.fixture(scope="session")
def api_server():
    proc = subprocess.Popen(
        [
            "uvicorn",
            "docs.api_example.server:app",
            "--host",
            "127.0.0.1",
            "--port",
            "8000",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    try:
        yield
    finally:
        proc.terminate()
        try:
            proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            proc.kill()

@pytest.mark.integration
@pytest.mark.parametrize(
    "notebook_path",
    list(NOTEBOOK_DIR.glob("*.ipynb")),
)
def test_example_notebook_runs_without_red_errors(api_server, notebook_path, tmp_path):
    with notebook_path.open() as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(
        timeout=600,
        kernel_name="python3",
    )
    ep.preprocess(nb, {"metadata": {"path": str(tmp_path)}})

    red_messages = []

    for cell in nb.cells:
        for output in cell.get("outputs", []):
            if output.get("output_type") != "stream":
                continue

            text = output.get("text", "")

            start = text.find(RED_ANSI)
            if start == -1:
                continue
            start_content = start + len(RED_ANSI)

            end = text.find(RESET_ANSI)
            content = text[start_content:end]
            red_messages.append(content)

    assert not red_messages, (
        f"Notebook {notebook_path} printed error messages in red:\n"
        + "\n---\n".join(red_messages)
    )



import subprocess
import requests

DOCKER_STATUS_GOOD_MESSAGE = "Docker is installed and the daemon is running."
DOCKER_NOT_INSTALLED_MESSAGE = "Docker is not installed or not available in the system PATH."
DOCKER_DAEMON_NOT_RUNNING_MESSAGE = "Docker daemon is not running. Please start the Docker service."

DOCKER_FORBIDDEN_FLAGS = [
    "--privileged",
    "--network=host",
    "--net=host",
]

class CommandResult:
    stdout: str = ""
    stderr: str = "" 

def validate_docker_args(args: str):
    tokens = args.split()
    for flag in DOCKER_FORBIDDEN_FLAGS:
        if flag in tokens:
            return False, f"Forbidden Docker flag detected: {flag}"
    return True, ""

class DockerBackend:
    def docker_exist(self):
        result = subprocess.run(["docker"], capture_output=True, text=True)
        return result.returncode == 0
    
    def docker_daemon_running(self):
        result = subprocess.run(["docker", "info"], capture_output=True, text=True)
        return result.returncode == 0
    
    def get_status(self):
        is_docker_available = self.docker_exist()
        if not is_docker_available:
            return False, DOCKER_NOT_INSTALLED_MESSAGE
        is_daemon_running = self.docker_daemon_running()
        if not is_daemon_running:
            return False, DOCKER_DAEMON_NOT_RUNNING_MESSAGE
        return True, DOCKER_STATUS_GOOD_MESSAGE
    
    def dockyter_command(self, cmd, args: str) -> CommandResult:
        result: CommandResult = CommandResult()

        ok, docker_status = self.get_status()
        if ok == False:
            result.stderr = docker_status
            return result

        ok, bad_flag_message = validate_docker_args(args)
        if ok == False: 
            result.stderr = bad_flag_message
            return result
        
        full_cmd = [
            "docker", "run", "--rm",
        ] + args.split() + [
            "bash", "-lc", cmd
        ]

        subprocessRes = subprocess.run(
            full_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,          
            encoding="utf-8",   
            errors="replace",   
        )

        result.stdout = subprocessRes.stdout
        result.stderr = subprocessRes.stderr

        return result

class APIBackend:
    def __init__(self, url):
        self.url = url.rstrip("/")

    def get_status(self):
        try:
            r = requests.get(f"{self.url}/health", timeout=10)
        except requests.RequestException as e:
            return False, f"API backend unreachable at {self.url}: {e}"
        if r.ok:
            return True, f"API backend reachable at {self.url}"
        return False, f"API backend error {r.status_code} at {self.url}"
        
    def dockyter_command(self, cmd, args: str) -> CommandResult:
        result: CommandResult = CommandResult()

        ok, api_status = self.get_status()
        if ok == False:
            result.stderr = api_status
            return result

        ok, bad_flag_message = validate_docker_args(args)
        if ok == False: 
            result.stderr = bad_flag_message
            return result
        
        body = {
            "cmd": cmd,
            "args": args,
        }

        try:
            r = requests.post(
                f"{self.url}/execute",
                json=body,
                timeout=10,
            )
            r.raise_for_status()
        except requests.RequestException as e:
            result.stderr = f"Error calling Dockyter API at {self.url}/execute: {e}"
            return result

        try:
            data = r.json()
        except ValueError:
            result.stderr = "Invalid JSON response from Dockyter API."
            return result

        result.stdout = data.get("stdout", "")
        result.stderr = data.get("stderr", "")

        return result
        



import dockyter.magics as magics_module
from dockyter.backend import CommandResult


class FakeShell:
    def __init__(self):
        self.calls = []
        self.system = self._default_system

    def _default_system(self, cmd):
        self.calls.append(cmd)

class FakeDockerBackend():
    def __init__(self):
        self.calls = []

    def get_status(self):
        return True, "Docker is installed and the daemon is running."

    def dockyter_command(self, cmd, args):
        res = CommandResult()
        res.stdout = f"Docker backend: {cmd} [{args}]"
        res.stderr = ""
        self.calls.append((cmd, args))
        return res

class FakeAPIBackend():
    def __init__(self, url):
        self.calls = []
        self.url = url

    def get_status(self):
         return True, f"API backend reachable at {self.url}"
    
    def dockyter_command(self, cmd, args):
        res = CommandResult()
        res.stdout = f"API backend: {cmd} [{args}]"
        res.stderr = ""
        self.calls.append((cmd, args))
        return res


def test_docker_backend_magic_switches_to_docker(monkeypatch, capsys):
    monkeypatch.setattr(magics_module, "DockerBackend", FakeDockerBackend)
    monkeypatch.setattr(magics_module, "APIBackend", FakeAPIBackend)

    shell = FakeShell()
    dockyter = magics_module.Dockyter(shell=shell)

    dockyter.docker_backend_magic("docker")
    assert isinstance(dockyter.backend, FakeDockerBackend)

    captured = capsys.readouterr().out
    assert "Switched backend to Docker (local daemon)." in captured
    assert "Dockyter backend type: Docker" in captured


def test_docker_backend_magic_switches_to_api(monkeypatch, capsys):
    monkeypatch.setattr(magics_module, "DockerBackend", FakeDockerBackend)
    monkeypatch.setattr(magics_module, "APIBackend", FakeAPIBackend)

    shell = FakeShell()
    dockyter = magics_module.Dockyter(shell=shell)

    dockyter.docker_backend_magic("api https://api.example.com")
    assert isinstance(dockyter.backend, FakeAPIBackend)
    assert dockyter.backend.url == "https://api.example.com"

    captured = capsys.readouterr().out
    assert "Switched backend to APIBackend (https://api.example.com)." in captured
    assert "Dockyter backend type: API" in captured


def test_docker_backend_magic_no_args(capsys):
    shell = FakeShell()
    dockyter = magics_module.Dockyter(shell=shell)

    dockyter.docker_backend_magic("")

    captured = capsys.readouterr().out
    assert "Usage: %docker_backend <docker|api> [api_url]" in captured


def test_docker_backend_magic_unknown_mode(capsys):
    shell = FakeShell()
    dockyter = magics_module.Dockyter(shell=shell)

    dockyter.docker_backend_magic("badmode")

    captured = capsys.readouterr().out
    assert "Unknown backend. Use docker or api." in captured


def test_docker_line_magic(monkeypatch, capsys):
    monkeypatch.setattr(magics_module, "DockerBackend", FakeDockerBackend)
    monkeypatch.setattr(magics_module, "APIBackend", FakeAPIBackend)

    shell = FakeShell()
    monkeypatch.setattr(magics_module, "get_ipython", lambda: shell)

    dockyter = magics_module.Dockyter(shell=shell)
    original_system = shell.system

    dockyter.docker_line_magic("ubuntu:latest -v /tmp:/tmp")
    assert dockyter.docker_args == "ubuntu:latest -v /tmp:/tmp"
    assert dockyter.docker_reroute_enabled is True
    assert shell.system == dockyter.docker_console  
    assert dockyter.original_system == original_system

    assert dockyter.backend.calls == [("echo 'Connected'", "ubuntu:latest -v /tmp:/tmp")]

    captured = capsys.readouterr().out
    assert "Connected" in captured


def test_docker_on_and_off_magics(monkeypatch):
    monkeypatch.setattr(magics_module, "DockerBackend", FakeDockerBackend)
    monkeypatch.setattr(magics_module, "APIBackend", FakeAPIBackend)

    shell = FakeShell()
    monkeypatch.setattr(magics_module, "get_ipython", lambda: shell)

    dockyter = magics_module.Dockyter(shell=shell)
    original_system = shell.system

    dockyter.docker_on_magic()
    assert dockyter.docker_reroute_enabled is True
    assert shell.system == dockyter.docker_console
    assert dockyter.original_system == original_system

    dockyter.docker_off_magic()
    assert dockyter.docker_reroute_enabled is False
    assert shell.system == original_system


def test_docker_cell_magic(monkeypatch, capsys):
    monkeypatch.setattr(magics_module, "DockerBackend", FakeDockerBackend)
    monkeypatch.setattr(magics_module, "APIBackend", FakeAPIBackend)

    shell = FakeShell()
    dockyter = magics_module.Dockyter(shell=shell)

    dockyter.docker_cell_magic("ubuntu:latest", "echo hello")
    assert dockyter.backend.calls == [("echo hello", "ubuntu:latest")]

    captured = capsys.readouterr().out
    assert "Docker backend: echo hello [ubuntu:latest]" in captured
